/* -----------------------------------------------------------------
;  Perform an EM iteration  
; ----------------------------------------------------------------- */
		
%macro EM_iter;
 
		*********************************************************
		*****	Multinomial Logit Model (k+1 th iteration)	*****
		*********************************************************;

       %IF &num_class > 1 %THEN %DO;
               
		PROC LOGISTIC DATA=work.mlogit;
	   		MODEL class (ref="&ref") = 
               %IF &num_covx > 0 %THEN %DO; 
                    &covx
               %END;
               / LINK=glogit;
	   		WEIGHT tau;
	   		OUTPUT OUT=work.logitpred predprobs=I;
	   		ODS OUTPUT ParameterEstimates=work.logitparam
		               FitStatistics = logit_fit;
                  format class class_f.;
	   	RUN;

        data work.logitparam;
	          set work.logitparam;
	          parameter = variable;
              * convert to numeric;
              class = 0 + response;             
	          keep parameter class estimate;
	       run;
        
      %END;
    	
		*********************************************************
		*****			WGEE Model (k+1 th iteration)		*****
		*********************************************************;
		
        %DO j = 1 %TO &num_feature; 
            %let Var = %scan(&features, &j);
		    %let Distr = %scan(&dist, &j);
            %WGEE(response = &Var, response_idx = &j, distrib = &Distr, corr=AR, tim = &time);
        %END;

		
		*********************************************************
		*****  	Data set of updated parameter estimates	    *****
		*********************************************************;
		TITLE;
		DATA work.param1;

           SET %IF &num_class > 1 %THEN %DO;
                  logitparam 
			   %END; 
		      %DO j = 1 %TO &num_feature;
      	        %DO g = 1 %TO &num_class;

                param&g&j
              %END;
           %END;
           ;   
           estimate1 = estimate;

			KEEP parameter class feature estimate1;
		RUN;

%mend EM_iter;
