/* -------------------------------------------------------------
;	Updated tau value at the (k+1)th iteration
;
;   Also, compute the model fit statistics.  Avoid using the QIC
;      and QICu outputs from PROC GENMOD, as this procedure does 
;      not compute the scale parameter correctly when weights are
;      present. 
; ------------------------------------------------------------- */
%macro update_tau;

PROC IML;
			
			*** Create two matrices:
				1. taunew which will contain the updated tau values for the multinomial
					   logit model (a vector)
				2. taunew2 which (when expanded appropriately) will contain the updated tau values for the WGEE (matrix);
			taunew = shape(.,&subjclass,1);

			taunew2 = shape(.,&num_subjects,&num_class);

            *** Create a matrix of cluster sizes of each feature;
			clustsize = shape(.,&num_subjects, &num_feature);

            *** Read in cluster sizes of each feature;
	        %DO j = 1 %TO &num_feature;   
	              USE work.clustsize&j;
		             read all var{size} into csize;
			      CLOSE clustsize&j;
				  clustsize[,&j] = csize;
			%END;
			
			*** read in latent class frequencies;
            USE work.logitpred;
				%DO g = 1 %TO &num_class;
					read all var{IP_&g} into pi&g;
				%END;
			CLOSE logitpred;

            *** Read in scale parameters;
	        %DO j = 1 %TO &num_feature;   
                %DO g = 1 %TO &num_class;
                    USE work.scale&g&j;
						read var{scale} into scale&g&j;
					CLOSE scale&g&j;
				%END;
			%END;

            *** Read in correlation matrices;
	        %DO j = 1 %TO &num_feature;   
                %DO g = 1 %TO &num_class;
                    USE work.corr&g&j;
						read all into wcorr&g&j;
					CLOSE corr&g&j;
				%END;
			%END;

            * convert from a macro string variable to a string variable;
            distribs = "&dist";

            * initialize the model fit statistics;
			entropy = 0;
			qic = 0;
			eqic = 0;
            %IF &num_class > 1 %THEN %DO;

			    USE work.logit_qic;
				   read all into logit_lik;
			    CLOSE logit_qic;
				qic = qic + logit_lik;
				eqic = eqic + logit_lik;

			%END;

             * compute projected LR and update tau, for each subject and latent class;
			%DO j = 1 %TO &num_feature;
                indx&j = 0;
			%END;
			
			* this loop cannot be a MACRO DO loop or else will run out of symbol space;

			DO i = 1 TO &num_subjects;

   				piLR = shape(0,1,&num_class);

			   *** note that log LR is a sum over the clinical features;

			   %DO j = 1 %TO &num_feature;

                 distrib = scan(distribs, &j); 
 		         size = clustsize[i,&j];

				 if size > 0 then do;
                   indxset = (indx&j + 1):(indx&j + size);
 				   indx&j = indx&j + size;
				   					
	 				%DO g = 1 %TO 1;

					   R_1 = wcorr&g&j[1:size,1:size];

                        USE work.resid&g&j;
						   read var{resid} into resid_1 POINT indxset;
						   read var{vmu} into vmu_1 POINT indxset;
						   read var{pred} into mu_1 POINT indxset;
                           read var{_weight} into weight_1 POINT indxset;
                           read var{_weight_tau} into weight_tau1 POINT indxset;
					    CLOSE resid&g&j;

					    A_1 = diag(vmu_1);

					*** Compute the variance-covariance matrix (feature, group, and subject specific);
					    V_1 = scale&g&j * A_1##(1/2) * R_1 * A_1##(1/2);
						Vinv_1 = inv(V_1);
                        weight_11 = diag(weight_1);
                        Vinv_1 = Vinv_1 * weight_11;

                        deviance_1 = 0;
                        piLR[1,1] = piLR[1,1] + deviance_1;

						* compute model fit statistics using weight_tau;
						y = resid_1 + mu_1;
                        if upcase(substr(distrib,1,3)) = "NOR" then 
						   qlik = -0.5*sum(weight_tau1#resid_1#resid_1)/scale&g&j;
                        if upcase(substr(distrib,1,3)) = "BIN" then 
						   qlik = sum(weight_tau1#(y#log(mu_1/(1-mu_1)) + log(1 - mu_1)));
	                    if upcase(substr(distrib,1,3)) = "POI" then do;
						   yint = round(y);
						   if mu_1 < 1E-12 then mu_1 = 1E-12;
						   * careful:  big mistake to divide log(fact(yint)) by the scale param;
                           qlik = sum(weight_tau1#(((yint#log(mu_1) - mu_1)/scale&g&j) - log(fact(yint))));
						end;
				
						/* extended qlik penalty: Nelder & Pregibon (1987) define the 
						   penalty slightly differently than Wang & Hin (2010) */
						* Wang & Hin (2010);
						* penalty = 0.5*sum(weight_tau1#log(vmu_1*scale&g&j));
						* Nelder & Pregibon (1987):  use V(y), which we can ignore,
						    instead of V(mu);
                        penalty = 0.5*sum(weight_tau1#log(scale&g&j));
                        eqlik = qlik - penalty;
						qic = qic - 2*qlik;
						eqic = eqic - 2*eqlik; 
                     %END;  

                     %IF &num_class > 1 %THEN %DO;
               
                      %DO g = 2 %TO &num_class;

                        R_g = wcorr&g&j[1:size,1:size];

						USE work.resid&g&j;
						   read var{resid} into resid_g POINT indxset;
                           read var{vmu} into vmu_g POINT indxset;
						   read var{pred} into mu_g POINT indxset;
						   read var{_weight} into weight_g POINT indxset;
                           read var{_weight_tau} into weight_taug POINT indxset;
					    CLOSE resid&g&j;

                        A_g = diag(vmu_g);

					*** Compute the variance-covariance matrix (feature, group, and subject specific);
					    V_g = scale&g&j * A_g##(1/2) * R_g * A_g##(1/2);
						Vinv_g = inv(V_g);
                        weight_gg = diag(weight_g);
                        Vinv_g = Vinv_g * weight_gg;

						
                    *** The contribution of the ith subject, gth class, jth feature to the deviance;
						deviance_g = 0.5 * (mu_g - mu_1)` * Vinv_1 * resid_1
									+ 0.5 * (mu_g - mu_1)` * Vinv_g * resid_g;

						piLR[1,&g] = piLR[1,&g] + (deviance_g - deviance_1);

						* compute model fit statistics using weight_tau;
						y = resid_g + mu_g;
                        if upcase(substr(distrib,1,3)) = "NOR" then 
						   qlik = -0.5*sum(weight_taug#resid_g#resid_g)/scale&g&j;
                        if upcase(substr(distrib,1,3)) = "BIN" then 
						   qlik = sum(weight_taug#(y#log(mu_g/(1-mu_g)) + log(1 - mu_g)));
	                    if upcase(substr(distrib,1,3)) = "POI" then do;
                           yint = round(y);
						   if mu_g < 1E-12 then mu_g = 1E-12; 
						   * careful:  big mistake to divide log(fact(yint)) by the scale param;
                           qlik = sum(weight_taug#(((yint#log(mu_g) - mu_g)/scale&g&j) - log(fact(yint))));
						end;

						/* extended qlik penalty: Nelder & Pregibon (1987) define the 
						   penalty slightly differently than Wang & Hin (2010) */
						* Wang & Hin (2010);
						* penalty = 0.5*sum(weight_taug#log(vmu_g*scale&g&j));
						* Nelder & Pregibon (1987):  use V(y), which we can ignore,
						    instead of V(mu);
                        penalty = 0.5*sum(weight_taug#log(scale&g&j));
                        eqlik = qlik - penalty;
						qic = qic - 2*qlik;
						eqic = eqic - 2*eqlik; 
                      %END;  
                     %END;  
                  END;
                
				%END;
								                               
                %DO g = 1 %TO &num_class; 
				    IF piLR[1,&g] > 40 THEN piLR[1,&g] = 40; 
				    IF piLR[1,&g] < -40 THEN piLR[1,&g] = -40;
                    piLR[1,&g] = pi&g[i] * exp(piLR[1,&g]);
				%END;
  
				*** This is the denominator for tau;
				denom = sum(piLR[1,]);

                %DO g = 1 %TO &num_class;

					tau_ig = piLR[1,&g]/denom;
					tauindex = (i-1) * &num_class + &g;
					taunew[tauindex,1] = tau_ig;
					taunew2[i,&g]=tau_ig;

					if tau_ig < 1E-10 then tau_ig = 1E-10;
					entropy = entropy - tau_ig * log(tau_ig);

				%END;

     		END;
	
		    * count the total number of interest parameters;
            USE work.param0;
		        read all var{estimate0} into estim;
  	        CLOSE param0;
            nparam = nrow(estim);	

			* Quasi-likelihood based AIC; 
            qaic = qic + 2*nparam;
			eqaic = eqic + 2*nparam;

           * Quasi-likelihood based BIC; 
            		
			qbic = qic + nparam*log(&num_subjects);
            eqbic = eqic + nparam*log(&num_subjects);

			* add classification entropy to BIC;

			ce_qbic = qbic + 2*entropy;
			ce_eqbic = eqbic + 2*entropy;

			* standardized entropy:  typically want this to be above 0.8 ;
			%IF &num_class = 1 %THEN %DO;
			   entropy_std = 1;
			%END;
			%ELSE %DO;
			   entropy_std = 1 - entropy/(&num_subjects*log(&num_class));
			%END;

            CREATE work.taunew FROM taunew[colname='taunew'];
			APPEND FROM taunew;

			CREATE work.taunew2 FROM taunew2;
			APPEND FROM taunew2;

			CREATE work.modelfit var {nparam qaic eqaic qbic eqbic ce_qbic ce_eqbic entropy entropy_std};
			APPEND;
			CLOSE modelfit;
		QUIT;

		*** grab the &subjectid variable so that we can merge the new tau
		    values with work.final;
		DATA work.taunew3;

			array tauvec[*] tau1 - tau&num_class;
          	array colvec[*] col1 - col&num_class;

		    MERGE work.logitpred (keep=&subjectid) work.taunew2;

 			DO g = 1 to &num_class;
			   tauvec[g] = colvec[g];
			END;

			DROP g col1 - col&num_class;

		RUN;


%mend;
