%macro WGEE(response =, response_idx = ,distrib = , corr=AR, tim = );

    %let _w = 1;
	%if &weight ne %then %let _w = &weight;

	%DO g = 1 %TO &num_class;

	* multiply tau by a weight to adjust for selective attrition;
       data work.final_temp;
	      set work.final;
		  _weight = 0 + &_w;
          _weight_tau = &_w * tau&g;

		  * avoid _weight_tau values of zero, since GENMOD would set both the raw residual and
		     the Pearson residual to missing;
		  if _weight_tau < 1E-10 then _weight_tau = 1E-10;

		  * derivative of the linear predictor;
	      eta_deriv = &tim;
		  keep &subjectid &response &tim eta_deriv _weight _weight_tau; 
	   run;

	   %LET poi_error = 0; 
	   %LET bin_error = 0;

		%*** Fit class specific GEE;
		%*** Note: For non-binary data, no longer need to ODS OUTPUT the scale parameter because I need to
			manually compute it to obtain the correct value;
		%IF &distrib ^= BIN %THEN %DO;
			PROC GENMOD DATA=work.final_temp;
				CLASS &subjectid;
				MODEL &response = &tim / dist = &distrib ;
				REPEATED SUBJECT = &subjectid / CORR = &corr corrw modelse;
				weight _weight_tau;
				OUTPUT OUT=work.resid&g&response_idx pred=pred resraw=resid	
				     reschi = pearson;
				ODS OUTPUT GEEEmpPEst= param_temp
						   GEEWCorr= work.corr&g&response_idx;
				TITLE "&g&response"; 
			RUN;

			* new March 2021: if needed, set the regression coefficients
			   in a Poisson log-linear model to a lower bound;
            %IF &distrib = POI %THEN %DO; 

			   %LET poi_intcpt = &boundcoef;
			   %LET poi_slope = &boundcoef;
               data param_temp&g&response_idx;
			      set param_temp;
				  pois_error = 0;
				  if estimate < &boundcoef then do;
                    pois_error = 1;
                    CALL SYMPUTX('poi_error', pois_error);
                  end;
				  if _n_ = 1 and estimate > &boundcoef then do;
				     pois_intcept = estimate;
                     CALL SYMPUTX('poi_intcpt', pois_intcept);
				  end;
                  else if _n_ = 2 and estimate > &boundcoef then do;
				     pois_slope = estimate;
                     CALL SYMPUTX('poi_slope', pois_slope);
				  end;
				run; 

				%IF &poi_error = 1 %THEN %DO; 

				* unfortunately, refitting the Poisson model initialized at the lower bound
				and limited to 0 or 1 iterations does not yield reasonable estimates in
				either PROC GENMOD or PROC GEE; 

           	* try an offset;
			data work.final_temp_poi;
			   set work.final_temp;
			   my_offset = &tim * &poi_slope;
			run;

            PROC GENMOD DATA=work.final_temp_poi;
				CLASS &subjectid;
				MODEL &response = / dist = &distrib noscale offset=my_offset 
                                         intercept=&poi_intcpt;
				REPEATED SUBJECT = &subjectid / CORR = &corr corrw modelse;
				weight _weight_tau;
				OUTPUT OUT=work.resid&g&response_idx pred=pred resraw=resid	
				     reschi = pearson;
				ODS OUTPUT GEEEmpPEst= param_temp
						   GEEWCorr= work.corr&g&response_idx;
				TITLE "&g&response"; 

         	RUN;

			data param_temp;
			   length parm $ 15;
			   set param_temp;
			   output;
			   parm = "&tim";
			   estimate = &poi_slope;
               output;
             run; 

			 * Pearson residuals and scale estimate are unreliable owing to underflow;
             data work.scale&g&response_idx;
			    length feature $ 15;
	   			scale_sqrt = 1; 
                scale = scale_sqrt**2;
                class = &g;
	            feature = "&response";
				KEEP class feature scale scale_sqrt;
			run;		

			*** compute V(mu) without using the Pearson residuals, and delete 
		     any missing observations;
		     DATA work.resid&g&response_idx;
		         SET work.resid&g&response_idx (KEEP = &subjectid &tim 
                          pred resid pearson eta_deriv _weight _weight_tau);
		         if missing(resid) then delete; 
				 vmu = pred; 
                 keep &subjectid &tim pred resid vmu eta_deriv _weight _weight_tau;
              RUN;

                %END;		
            %END;

			%IF &poi_error = 0 %THEN %DO; 

			*** Correct for the incorrect scale estimate;
			PROC MEANS DATA=work.resid&g&response_idx noprint;
	   			var pearson _weight_tau;
	   			output out=work.sumresid uss=uss_res uss_tau sum=sum_res sum_tau;
			run;

			*** Note: scale_param is the corrected scale parameter corresponding to the square root
				of the dispersion parameter.  Scale is the scale parameter used to compute the variance-covariance
				matrix;
			DATA work.scale&g&response_idx;
			    length feature $ 15;
	   			SET work.sumresid;
	   			scale_sqrt = sqrt(uss_res/sum_tau);
				scale = scale_sqrt**2;
                class = &g;
	            feature = "&response";
				KEEP class feature scale scale_sqrt;
			run;	

			*** compute V(mu) using the Pearson residuals, and delete 
		     any missing observations;
		DATA work.resid&g&response_idx;
		   SET work.resid&g&response_idx (KEEP = &subjectid &tim 
                          pred resid pearson eta_deriv _weight _weight_tau);
		   if missing(resid) then delete; 

		   vmu = sqrt(_weight_tau) * resid/pearson;
		   vmu = vmu * vmu;
		   keep &subjectid &tim pred resid vmu eta_deriv _weight _weight_tau;
        RUN;

			%END;
        %END;
     %ELSE %DO;
		*** If the data is binary then use the descending option to model the
		correct probability;
		
			PROC GENMOD DATA=work.final_temp descending;
				CLASS &subjectid;
				MODEL &response = &tim / dist = &distrib noscale;
				REPEATED SUBJECT = &subjectid / CORR = &corr corrw modelse;
				weight _weight_tau;
				OUTPUT OUT=work.resid&g&response_idx pred=pred resraw=resid	
				     reschi = pearson;
				ODS OUTPUT GEEEmpPEst= param_temp
						   GEEWCorr= work.corr&g&response_idx;
				TITLE "&g&response";
			RUN;
 
 			DATA work.scale&g&response_idx;
			    length feature $ 15;
				scale_sqrt = 1;
				scale = scale_sqrt**2;
                class = &g;
	            feature = "&response";
				KEEP class feature scale_sqrt scale;
			RUN;
	
            * new March 2021: if needed, set the regression coefficients
			   in a binomial logistic model to a lower bound;
            %LET bin_intcpt = &boundcoef;
			%LET bin_slope = &boundcoef;

               data param_temp&g&response_idx;
			      set param_temp;
				  bern_error = 0;
				  if estimate < &boundcoef then do;
                    bern_error = 1;
                    CALL SYMPUTX('bin_error', bern_error);
                  end;
				  if _n_ = 1 and estimate > &boundcoef then do;
				     bern_intcept = estimate;
                     CALL SYMPUTX('bin_intcpt', bern_intcept);
				  end;
                  else if _n_ = 2 and estimate > &boundcoef then do;
				     bern_slope = estimate;
                     CALL SYMPUTX('bin_slope', bern_slope);
				  end;
				run; 

				%IF &bin_error = 1 %THEN %DO;  

				* try an offset;
			      data work.final_temp_bin;
        			   set work.final_temp;
		         	   my_offset = &tim * &bin_slope;
			      run;

            PROC GENMOD DATA=work.final_temp_bin descending;
				CLASS &subjectid;
				MODEL &response = / dist = &distrib noscale offset=my_offset 
                                         intercept=&bin_intcpt;
				REPEATED SUBJECT = &subjectid / CORR = &corr corrw modelse;
				weight _weight_tau;
				OUTPUT OUT=work.resid&g&response_idx pred=pred resraw=resid	
				     reschi = pearson;
				ODS OUTPUT GEEEmpPEst= param_temp
						   GEEWCorr= work.corr&g&response_idx;
				TITLE "&g&response"; 
         	RUN;
            
			data param_temp;
			   length parm $ 15;
			   set param_temp;
			   output;
			   parm = "&tim";
			   estimate = &bin_slope;
               output;
             run; 

			 %END;

			 *** compute Binomial V(mu) without using (possibly unstable or missing) Pearson residuals, and delete 
		     any missing observations;
		     DATA work.resid&g&response_idx;
		         SET work.resid&g&response_idx (KEEP = &subjectid &tim 
                          pred resid pearson eta_deriv _weight _weight_tau);
		         if missing(resid) then delete; 
				 vmu = pred * (1 - pred); 
                 keep &subjectid &tim pred resid vmu eta_deriv _weight _weight_tau;
              RUN;

		%END; 

        data work.param&g&response_idx;
		   length feature $ 15;
	       set param_temp;
           class = &g;
	       feature = "&response";
		   parameter = parm;
	       keep parameter class feature estimate;
	    run;
  		
	%END;
	
%mend WGEE;
